package com.gxzc.zen.orm.aspect

import com.gxzc.zen.orm.DynamicMultipleDataSource
import com.gxzc.zen.orm.annotation.DynamicDataSource
import com.gxzc.zen.orm.contants.DSKey
import org.aspectj.lang.JoinPoint
import org.aspectj.lang.ProceedingJoinPoint
import org.aspectj.lang.annotation.Around
import org.aspectj.lang.annotation.Aspect
import org.aspectj.lang.annotation.Before
import org.aspectj.lang.annotation.Pointcut
import org.aspectj.lang.reflect.MethodSignature
import org.slf4j.LoggerFactory
import org.springframework.core.annotation.Order
import org.springframework.stereotype.Component
import java.lang.reflect.Method

/**
 *
 * @author NorthLan at 2018/1/29
 */
@Suppress("unused")
@Aspect
@Order(-100) // 保证先于事务执行
@Component
class DataSourceSwitchAspect {
    companion object {
        private val logger = LoggerFactory.getLogger(DataSourceSwitchAspect::class.java)
    }

    private var isAnnotationAspect = false

    @Pointcut("execution(* com.gxzc.zen.api..*Service.*(..))")
    fun zenServicePointCut() {
    }

    @Pointcut("execution(* com.baomidou.mybatisplus..*Service.*(..))")
    fun mpServicePointCut() {
    }

    @Pointcut("execution(* com.baomidou.mybatisplus..*Mapper.*(..))")
    fun mpMapperPointCut() {
    }

    @Pointcut("@annotation(com.gxzc.zen.orm.annotation.DynamicDataSource)")
    fun annotationCut() {
    }

    @Around("annotationCut()")
    fun annotationAround(joinPoint: ProceedingJoinPoint): Any? {
        isAnnotationAspect = true
        val method = (joinPoint.signature as MethodSignature).method
        setDataSource(method, DSKey.DSKEY_SYS)
        try {
            return joinPoint.proceed()
        } finally {
            isAnnotationAspect = false
            DynamicMultipleDataSource.clear()
        }
    }

    @Before("zenServicePointCut() || mpServicePointCut() || mpMapperPointCut()")
    fun dynamicDataSource(joinPoint: JoinPoint) {
        if (isAnnotationAspect) {
            return
        }
        val target = joinPoint.target

        val packageName: String = try {
            val h = target::class.java.superclass.getDeclaredField("h")
            h.isAccessible = true
            val mapperProxy = h.get(target)

            val mapperInterface = mapperProxy.javaClass.getDeclaredField("mapperInterface")
            mapperInterface.isAccessible = true

            val clazz = mapperInterface.get(mapperProxy) as Class<*>

            clazz.`package`.name
        } catch (e: Exception) {
            logger.debug("get packageName of MapperProxy error, ignore if nothing happened!")
            logger.debug("set packageName of target...")
            target::class.java.`package`.name
        }

        matchDataSource((joinPoint.signature as MethodSignature).method, packageName)
    }

    private fun matchDataSource(method: Method, packageName: String) {
        DSKey.values()
                .filter { it.pkgName in packageName }
                .forEach { setDataSource(method, it.dsKey) }
    }

    /**
     * 设置数据源key
     */
    private fun setDataSource(method: Method, defaultKey: String) {
        val dynamicDataSource: DynamicDataSource? = method.getAnnotation(DynamicDataSource::class.java)
        DynamicMultipleDataSource.setDataSource(dynamicDataSource?.value ?: defaultKey)
    }
}